#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : completions
# @Time         : 2023/12/19 16:38
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from meutils.serving.fastapi.dependencies.auth import get_bearer_token, HTTPAuthorizationCredentials

from fastapi import APIRouter, File, UploadFile, Query, Form, Depends, Request

from sse_starlette import EventSourceResponse

from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk

from chatllm.llmchain.completions import github_copilot
from chatllm.llmchain.completions import moonshot_kimi
from chatllm.llmchain.completions import deepseek

from chatllm.schemas.openai_api_protocol import ChatCompletionRequest, UsageInfo

router = APIRouter()

ChatCompletionResponse = Union[ChatCompletion, List[ChatCompletionChunk]]


@router.post("/chat/completions")
def chat_completions(
    request: ChatCompletionRequest,
    auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token)
):
    logger.debug(request)

    api_key = auth and auth.credentials or None
    logger.debug(api_key)

    model = request.model.strip().lower()
    stream = request.stream
    data = request.model_dump()

    if model.startswith(('kimi', 'moonshot')):
        if any(i in model for i in ('web', 'search', 'net')):
            data['use_search'] = True  # 联网模型
        # state_file = "/www/0_apps/cookies/kimi_cookies.json"
        # api_key = Path(state_file).exists() and state_file or None

        completions = moonshot_kimi.Completions(api_key=api_key)

    elif model.startswith(('deepseek',)):
        completions = deepseek.Completions(api_key=api_key)

    else:  # todo: 兜底
        completions = github_copilot.Completions(api_key=api_key)

    response: ChatCompletionResponse = completions.create(**data)
    if stream:
        # generator = map(lambda chunk: chunk.model_dump_json(), response)
        generator = (chunk.model_dump_json() for chunk in response)

        return EventSourceResponse(generator, ping=10000)  # todo: 耗时监控

    return response


if __name__ == '__main__':
    from meutils.serving.fastapi import App

    app = App()

    app.include_router(router, '/v1')

    app.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : openai_types
# @Time         : 2023/12/19 09:46
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import os

from meutils.pipe import *
from openai.types.chat.completion_create_params import CompletionCreateParams
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion import ChatCompletion

COMPLETION_ERROR = os.getenv("COMPLETION_ERROR", "请稍后再试！！！")

chat_completion_error = {
    "id": "chatcmpl-",
    "object": "chat.completion",
    "created": 0,
    "model": "LLM",
    "choices": [{"message": {"role": "assistant", "content": COMPLETION_ERROR}, "index": 0, "finish_reason": "stop"}]
}

chat_completion_chunk_error = {
    "id": "chatcmpl-",
    "object": "chat.completion.chunk",
    "created": 0,
    "model": "LLM",
    "choices": [{"delta": {"role": "assistant", "content": COMPLETION_ERROR}, "index": 0, "finish_reason": "stop"}]
}

chat_completion_error = ChatCompletion.model_validate(chat_completion_error)
chat_completion_chunk_error = ChatCompletionChunk.model_validate(chat_completion_chunk_error)


class ChatCompletionRequest(BaseModel):
    request: CompletionCreateParams = {}
    headers: dict = {}  # HTTPException(status.HTTP_401_UNAUTHORIZED, "Token is wrong!")


if __name__ == '__main__':
    request = {"request": {"stream": True, "model": "gpt-3.5-turbo", "messages": [{"role": "user", "content": "你好"}]}}

    print(ChatCompletionRequest.model_validate(request))

    print(chat_completion_chunk_error)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : generativeai
# @Time         : 2023/12/14 13:32
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://github.com/google/generative-ai-docs/blob/main/site/en/tutorials/python_quickstart.ipynb

from meutils.pipe import *
from meutils.cache_utils import ttl_cache
from meutils.decorators.retry import retrying

from chatllm.llmchain.utils import tiktoken_encoder
from chatllm.schemas.kimi.protocol import EventData
from chatllm.llmchain.completions import github_copilot

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

import google.generativeai as genai
from google.generativeai import GenerativeModel

# Set up the model
generation_config = {
    "temperature": 0.9,
    "top_p": 1,
    "top_k": 1,
    "max_output_tokens": 2048,
}

safety_settings = [
    {
        "category": "HARM_CATEGORY_HARASSMENT",
        "threshold": "BLOCK_ONLY_HIGH"
    },
    {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "threshold": "BLOCK_ONLY_HIGH"
    },
    {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "threshold": "BLOCK_ONLY_HIGH"
    },
    {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "threshold": "BLOCK_ONLY_HIGH"
    }
]


class Completions(github_copilot.Completions):

    def __init__(self, **client_params):
        self.client_params = client_params
        api_key = self.client_params.get('api_key')
        genai.configure(api_key=api_key or os.getenv('GOOGLE_API_KEY'))

    def create(self, messages: Union[str, List[Dict[str, Any]]], **kwargs):  # ChatCompletionRequest: 定义请求体

        data = {
            "model": 'gpt-4',
            "messages": messages if isinstance(messages, list) else [{"role": "user", "content": messages}],
            **kwargs
        }

        # logger.debug(data)

        if data.get('stream'):
            interval = 0.01
            return self.smooth_stream(interval=interval, **data)
        else:
            return self._create(**data)

    def _create(self, **data):
        response = self._post(**data)

        chunk_id = f"chatcmpl-{uuid.uuid1()}"
        created = int(time.time())
        model = data.get('model', 'kimi')  # "kimi-clk4da83qff43om28p80|clk4da83qff43om28p80"
        finish_reason = "stop"

        chunk = {
            'id': chunk_id,
            'choices': [
                {
                    'finish_reason': finish_reason,
                    'index': 0,
                    'message': {
                        'content': response.text,
                        'role': 'assistant',
                        'function_call': None,
                        'tool_calls': None},
                }
            ],
            'created': created,
            'model': model,
            'object': 'chat.completion',
            'system_fingerprint': None,
            'usage': {
                'completion_tokens': 188,
                'prompt_tokens': 1115,
                'total_tokens': 1303
            }
        }

        return ChatCompletion.model_validate(chunk)

    def _stream_create(self, **data):
        response = self._post(**data)

        chunk_id = f"chatcmpl-{uuid.uuid1()}"
        created = int(time.time())
        model = data.get('model', 'kimi')  # "kimi-clk4da83qff43om28p80|clk4da83qff43om28p80"
        finish_reason = None

        for chunk in response:
            chunk = {
                'id': chunk_id,
                'choices': [
                    {
                        'delta': {
                            'content': chunk.text,
                            'function_call': None,
                            'role': 'assistant',
                            'tool_calls': None
                        },
                        'finish_reason': finish_reason,
                        'index': 0
                    }
                ],
                'created': created,
                'model': model,
                'object': 'chat.completion.chunk',
                'system_fingerprint': None
            }
            chunk = ChatCompletionChunk.model_validate(chunk)
            # logger.debug(chunk)

            yield chunk
            # chunk.choices[0].finish_reason = 'stop' # 怎么判断
            # yield chunk

    def _post(self, **data):
        model_name = data.pop("model", "gemini-pro")
        messages = json.dumps(data.pop("messages"))  # 模拟多轮对话
        stream = data.pop("stream", False)

        generation_config = data
        # generation_config = {
        #     "temperature": 0.9,
        #     "top_p": 1,
        #     "top_k": 1,
        #     "max_output_tokens": 2048,
        # }
        model = genai.GenerativeModel(
            model_name=model_name,
            generation_config=generation_config,
            safety_settings=safety_settings
        )
        return model.generate_content(messages, stream=stream)


if __name__ == '__main__':
    messages = [
        {'role': 'system', 'content': "你是个python专家"}, {'role': 'user', 'content': '你是谁'},
        {'role': 'assistant', 'content': '我是你的智能助手。'}, {'role': 'user', 'content': '我有点难过'},
        {'role': 'assistant', 'content': '你想让我帮你吗？'}, {'role': 'user', 'content': '帮我分担一下吧'}
    ]



from chatllm.llmchain.decorators import llm_stream
from meutils.pipe import *
import streamlit as st

from meutils.pipe import *
from chatllm.llmchain.llms import SparkBot

from langchain.chains import LLMChain
from langchain.prompts import ChatPromptTemplate

first_prompt = ChatPromptTemplate.from_template("{q}")

llm = SparkBot(streaming=True)
chain = LLMChain(llm=llm, prompt=first_prompt)


def on_btn_click():
    del st.session_state.messages


user_prompt = "<|User|>:{user}<eoh>\n"
robot_prompt = "<|Bot|>:{robot}<eoa>\n"
cur_query_prompt = "<|User|>:{user}<eoh>\n<|Bot|>:"


def combine_history(prompt):
    messages = st.session_state.messages
    total_prompt = ""
    for message in messages:
        cur_content = message["content"]
        if message["role"] == "user":
            cur_prompt = user_prompt.replace("{user}", cur_content)
        elif message["role"] == "robot":
            cur_prompt = robot_prompt.replace("{robot}", cur_content)
        else:
            raise RuntimeError
        total_prompt += cur_prompt
    total_prompt = total_prompt + cur_query_prompt.replace("{user}", prompt)
    return total_prompt


def main():
    # torch.cuda.empty_cache()

    user_avator = "user.png"
    robot_avator = "robot.png"

    st.title("金融大模型")

    # Initialize chat history
    if "messages" not in st.session_state:
        st.session_state.messages = []

    # Display chat messages from history on app rerun
    for message in st.session_state.messages:
        with st.chat_message(message["role"], avatar=message.get("avatar")):
            st.markdown(message["content"])

    # Accept user input
    if prompt := st.chat_input("What is up?"):
        # Display user message in chat message container
        with st.chat_message("user", avatar=user_avator):
            st.markdown(prompt)
        real_prompt = combine_history(prompt)
        print(f"real_prompt: {real_prompt}")

        # Add user message to chat history
        st.session_state.messages.append({"role": "user", "content": prompt, "avatar": user_avator})

        with st.chat_message("robot", avatar=robot_avator):
            message_placeholder = st.empty()
            cur_response = ''
            for cur_response_ in llm_stream(chain.run)(real_prompt):
                # Display robot response in chat message container
                cur_response += cur_response_
                print(cur_response)

                message_placeholder.markdown(cur_response + "▌")

            message_placeholder.markdown(cur_response)
        # Add robot response to chat history
        st.session_state.messages.append({"role": "robot", "content": cur_response, "avatar": robot_avator})


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : VectorRecordManager
# @Time         : 2023/9/12 13:40
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://python.langchain.com/docs/modules/data_connection/indexing#using-with-loaders

from langchain.embeddings import OpenAIEmbeddings
from langchain.indexes import SQLRecordManager, index
from langchain.schema import Document
from langchain.vectorstores import Chroma, VectorStore
from langchain.document_loaders.base import BaseLoader

from meutils.pipe import *
from chatllm.llmchain.vectorstores import ElasticsearchStore


# 混合检索 https://zhuanlan.zhihu.com/p/665097446
class VectorRecordManager(object):
    """
    增量更新向量
        manager.vectorstore.similarity_search(
        'doc',
        filter=[{'term': {'metadata.source': 'unknown'}}]
    )
    """

    def __init__(
        self, collection_name="test_index",
        vectorstore: Optional[VectorStore] = None,
        db_url: Optional[str] = None,
    ):
        """

        :param collection_name:
        :param vectorstore:
            # 本地
            vectorstore = Chroma(collection_name=collection_name, embedding_function=embedding)

        :param db_url:
            # 默认在 HOME_CACHE
            f"sqlite:///{HOME_CACHE}/chatllm/vector_record_manager.sql"

            "sqlite:///chatllm_vector_record_manager_cache.sql"

        """
        self.collection_name = collection_name
        self.vectorstore = vectorstore or ElasticsearchStore(
            embedding=OpenAIEmbeddings(),
            index_name=self.collection_name,  # 同一模型的embedding
            es_url=os.getenv('ES_URL'),
            es_user=os.getenv('ES_USER'),
            es_password=os.getenv('ES_PASSWORD'),
        )
        namespace = f"{self.vectorstore.__class__.__name__}/{collection_name}"
        db_url = db_url or f"sqlite:///{HOME_CACHE / 'chatllm/vector_record_manager.sql'}"

        self.record_manager = SQLRecordManager(namespace, db_url=db_url)
        self.record_manager.create_schema()

    def update(
        self,
        docs_source: Union[List[str], BaseLoader, Iterable[Document]],
        cleanup: Literal["incremental", "full", None] = "incremental",
        source_id_key: Union[str, Callable[[Document], str], None] = "source",
    ):
        """

        :param docs_source:
        :param cleanup:
        :param source_id_key: 根据 metadata 信息去重
        :return:
        """
        if isinstance(docs_source, List) and isinstance(docs_source[0], str):
            docs_source = [Document(page_content=text, metadata={"source": 'unknown'}) for text in docs_source]

        return index(docs_source, self.record_manager, self.vectorstore, cleanup=cleanup, source_id_key=source_id_key)

    def clear(self):
        return index([], self.record_manager, self.vectorstore, cleanup="full", source_id_key="source")


if __name__ == '__main__':
    doc1 = Document(page_content="kitty", metadata={"source": "kitty.txt"})
    doc2 = Document(page_content="doggy", metadata={"source": "doggy.txt"})
    with timer():
        manager = VectorRecordManager()
    # print(manager.clear())
    # print(manager.update([doc1, doc2] * 30))

    # 检索
    # filter_term = [{"term": {"metadata.source": "kitty.txt"}}]
    # print(manager.vectorstore.similarity_search(query='kitty', filter=filter_term, threshold=0.5))

    # filter_term = [{"term": {"metadata.source": "同程商旅Q&A.pdf"}}]
    # filter_term = [{"match": {"metadata.source": "同程商旅Q&A.pdf"}}]
    #
    # _ = (
    #     VectorRecordManager(collection_name='zhihe_test')
    #     .vectorstore
    #     .similarity_search(
    #         query='同程商旅Q&A通用版 商旅大客户服务组 陶璐',
    #         filter=filter_term,
    #         threshold=0.5
    #     )
    # )
    # print(_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : base
# @Time         : 2023/7/5 09:32
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *

from langchain.vectorstores.base import Document, CallbackManagerForRetrieverRun
from langchain.vectorstores.base import VectorStoreRetriever as _VectorStoreRetriever

class VectorStoreRetriever(_VectorStoreRetriever):

    def _get_relevant_documents(
        self, query: str, *, run_manager: Optional[CallbackManagerForRetrieverRun]
    ) -> List[Document]:
        if self.search_type == "similarity":
            docs = self.vectorstore.similarity_search(query, **self.search_kwargs)
        elif self.search_type == "similarity_score_threshold":
            docs_and_similarities = (
                self.vectorstore.similarity_search_with_relevance_scores(
                    query, **self.search_kwargs
                )
            )

            # docs = [doc for doc, _ in docs_and_similarities]
            docs = []
            for doc, score in docs_and_similarities:
                logger.info(docs_and_similarities)
                doc.metadata['similarity_score'] = score
                docs.append(doc)

        elif self.search_type == "mmr":
            docs = self.vectorstore.max_marginal_relevance_search(
                query, **self.search_kwargs
            )
        else:
            raise ValueError(f"search_type of {self.search_type} not allowed.")
        return docs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : dify
# @Time         : 2023/9/11 10:35
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from chatllm.schemas.openai_api_protocol import *
from meutils.pipe import *


class Message(BaseModel):
    event: Literal['message', 'message_end']
    id: str
    task_id: str
    conversation_id: str
    created_at: Optional[int] = Field(default_factory=lambda: int(time.time()))
    answer: Optional[str] = ''


class DifyCompletion(object):

    @classmethod
    def create(
        cls,
        messages: List[Dict[str, Any]],  # [{'role': 'user', 'content': '讲个故事'}]
        **kwargs,
    ):
        return cls.stream_create(messages[0]['content'], kwargs.pop('api_key', ''), **kwargs)

    @staticmethod
    def stream_create(content, api_key, **kwargs):
        headers = {
            'Authorization': f'Bearer {api_key}'
        }
        data = {
            "inputs": {},
            "query": content,
            "response_mode": "streaming",
            "conversation_id": "",
            "user": "USER"
        }
        response = requests.post('https://api.dify.ai/v1/chat-messages', json=data, headers=headers, stream=True)

        stream_resp = {}
        for chunk in response.iter_lines(decode_unicode=True):
            _ = chunk.split('data:')[-1].strip()
            if _:
                chunk = Message.parse_raw(_).answer

                ####################################################################################
                delta = DeltaMessage(role='assistant', content=chunk)

                choice_data = ChatCompletionResponseStreamChoice(
                    index=0,
                    delta=delta,
                    finish_reason=None,  # 最后一个是stop
                )

                stream_resp = ChatCompletionStreamResponse(
                    choices=[choice_data],
                    model=kwargs.get('model', 'dify-app'),
                ).dict()

                ####################################################################################
                yield stream_resp
        stream_resp['choices'][0]['finish_reason'] = 'stop'
        yield stream_resp


if __name__ == '__main__':
    api_key = 'app-n75siugXhOhgosA6YkLpTH5X'
    for i in DifyCompletion.create([{'role': 'user', 'content': '讲个故事'}], api_key=api_key):
        print(i['choices'][0]['delta']['content'], end='')
